/*
** Copyright 2017 Double Precision, Inc.
** See COPYING for distribution information.
*/
#include "libcxxw_config.h"
#include "combobox/custom_combobox_container_impl.H"
#include "combobox/custom_comboboxlayoutmanager.H"
#include "combobox/custom_combobox_popup_container_impl.H"
#include "combobox/combobox_button_impl.H"
#include "peepholed_toplevel_listcontainer/create_popup.H"
#include "listlayoutmanager/list_element_impl.H"
#include "listlayoutmanager/listlayoutstyle_impl.H"

#include "focus/focusframecontainer_element.H"
#include "focus/focusable_element.H"

#include "x/w/focusable_container.H"
#include "x/w/key_event.H"
#include "image_button_internal.H"
#include "icon_images_vector_element.H"
#include "hotspot_element.H"
#include "icon.H"
#include "generic_window_handler.H"
#include "capturefactory.H"
#include "run_as.H"

#include <x/weakcapture.H>
#include <x/visitor.H>
#include <X11/keysym.h>

LIBCXXW_NAMESPACE_START

custom_comboboxlayoutmanagerObj
::custom_comboboxlayoutmanagerObj(const ref<implObj> &impl,
				  const ref<listlayoutmanagerObj::implObj>
				  &list_layout_impl)
	// Surprise! Our implementation object is attached to the combobox
	// container, but our superclass is a listlayoutmanager whose
	// implementation object is the popup's listlayoutmanager implementation
	// object.
	//
	// It is expected that most of the activity here will impact the
	// combobox popup, so naturally when we go out of scope and get
	// destroyed it'll be the popup's container that'll get tickled for
	// recalculation.
	: listlayoutmanagerObj(list_layout_impl),
	  impl(impl)
{
}

custom_comboboxlayoutmanagerObj::~custom_comboboxlayoutmanagerObj()=default;

element custom_comboboxlayoutmanagerObj::current_selection()
{
	// The current selection element is always position (0, 0)
	// in the internally-managed grid.

	return impl->get(0, 0);
}

const_element custom_comboboxlayoutmanagerObj::current_selection() const
{
	// The current selection element is always position (0, 0)
	// in the internally-managed grid.

	return impl->get(0, 0);
}

/////////////////////////////////////////////////////////////////////////////
//
// Combobox button.

typedef focusframecontainer_elementObj<container_elementObj<child_elementObj>
				       > combobox_button_focusframe_container_t;


// Create the combobox button.

// The factory where the combobox button gets created gets passed in.

static inline auto
create_combobox_button(const ref<containerObj::implObj> &parent_container,
		       const auto &popup_handler)
{
	// The actual element that will go into this factory will be the
	// focus frame, with the button inside it. First, construct
	// the focus frame implementation object.

	auto cbfc=ref<combobox_button_focusframe_container_t>
		::create("comboboxbuttonfocusoff_border",
			 "comboboxbuttonfocuson_border",
			 parent_container,
			 child_element_init_params{"focusframe@libcxx.com"});

	// The focus frame implementation object is the parent of the
	// combobox button. Create its implementatio n button.

	auto &d=cbfc->get_element_impl().get_window_handler();

	auto icon1=d.create_icon_pixels("scroll-down1", render_repeat::none,
					0, 0,
					icon_scale::nomore);
	auto icon2=d.create_icon_pixels("scroll-down2", render_repeat::none,
					0, 0,
					icon_scale::nomore);

	auto image_button_internal_impl=
		ref<combobox_button_implObj>
		::create(cbfc,
			 std::vector<icon>{ icon1, icon2 },
			 popup_handler);

	// We can now create the focusframe public object.

	auto ff=focusframecontainer::create(cbfc, image_button_internal_impl);

	// The focus frame's factory, where the focusable element, the
	// image button, gets created.
	auto focusframe_factory=ff->set_focusable();

	// Create the "public" object, show() it, and tell the focus frame:
	// here's what you hafe inside you.
	auto combobox_button=image_button_internal
		::create(image_button_internal_impl);

	combobox_button->show();

	focusframe_factory->created_internally(combobox_button);

	ff->show();

	ff->label_for(ff); // Make clicks on the focusframe work.

	return std::make_tuple(ff, combobox_button);
}

/////////////////////////////////////////////////////////////////////////////
//
// Internal object that collects keystrokes when the input focus is on the
// current selection element, for the lookup callback.

class LIBCXX_HIDDEN lookup_collectorObj : virtual public obj {

 public:

	std::u32string buffer;

	lookup_collectorObj()=default;
	~lookup_collectorObj()=default;

	// Called to process a key event.

	// Collect the key event into the buffer, and invoke the
	// combo-box's search callback.
	//
	// This is always invoked from the connection thread.

	inline bool process(const all_key_events_t &e,
			    bool activated,
			    const custom_combobox_selection_search_t
			    &search_func,
			    const element &current_selection,
			    const custom_comboboxlayoutmanager &lm,
			    const busy &mcguffin)
	{
		size_t i=0;

		list_lock lock{lm};

		if (!std::visit(visitor{
		    [&](const key_event *ke)
		    {
			if (!ke->notspecial())
				return false;

			switch (ke->keysym) {
			case XK_Delete:
			case XK_KP_Delete:
				if (activated)
					buffer.clear();
				break;
			case XK_Up:
			case XK_KP_Up:
				if (activated)
				{
					auto selected=lm->selected();

					i=selected ? *selected:lm->size();

					while (i)
					{
						--i;
						if (!lm->enabled(i))
							continue;

						lm->autoselect(i);
						break;
					}
				}
				buffer.clear();
				// Set "activated" to false in order to bail
				// out, below, instead of calling search_func().
				// We handled everything here.
				activated=false;
				return true;
			case XK_Down:
			case XK_KP_Down:
				if (activated)
				{
					auto selected=lm->selected();

					i=selected ? *selected + 1 : 0;

					auto n=lm->size();

					while (i<n)
					{
						if (lm->enabled(i))
						{
							lm->autoselect(i);
							break;
						}
						++i;
					}
				}
				buffer.clear();
				// Set "activated" to false in order to bail
				// out, below, instead of calling search_func().
				// We handled everything here.
				activated=false;
				return true;
			default:

				if (!ke->unicode)
					return false;

				if (ke->unicode == '\n')
				{
					if (!activated)
						break;
					// Get the current selection, and
					// start the search on the next list item.

					auto selected=lm->selected();

					if (selected)
					{
						i=selected.value();
						++i;
					}
				}
				else
				{
					if (ke->unicode < ' ')
					{
						return false;
					}
					if (!activated)
						break;
					buffer.push_back(ke->unicode);
				}
			}
			return true;
		    },
		    [&](const std::u32string_view *str)
		    {
			    buffer += *str;
			    return true;
		    }
				}, e))
		{
			return false;
		}

		if (!activated)
			return true;

		if (search_func)
			search_func({lock, lm, buffer, i, current_selection,
						mcguffin});

		return true;
	}
};

typedef ref<lookup_collectorObj> lookup_collector;

/////////////////////////////////////////////////////////////////////////////

new_custom_comboboxlayoutmanager
::new_custom_comboboxlayoutmanager(const custom_combobox_selection_factory_t
				   &selection_factory)
	: selection_factory(selection_factory),
	  selection_changed([]
			    (const auto &ignore)
			    {
			    })
{
}

new_custom_comboboxlayoutmanager::~new_custom_comboboxlayoutmanager()=default;

custom_combobox_selection_changed_t
new_custom_comboboxlayoutmanager::get_selection_changed() const
{
	return selection_changed;
}

focusable_container new_custom_comboboxlayoutmanager
::create(const ref<containerObj::implObj> &parent) const
{
	// Start by creating the popup first.

	new_listlayoutmanager style{highlighted_list};

	style.background_color="combobox_background_color";
	style.selected_color="combobox_selected_color";
	style.highlighted_color="combobox_highlighted_color";
	style.current_color="combobox_current_color";

	custom_combobox_popup_containerptr popup_containerptr;

	auto [combobox_popup, popup_handler]=
		create_peepholed_toplevel_listcontainer_popup
		({
			ref<elementObj::implObj>(&parent->get_element_impl()),
				"combo,popup_menu,dropdown_menu",
				"combobox",
				"combobox_border",

				// We're about to create the combobox container,
				// with nesting_level of parent+1
				//
				// The current selection element in the combox
				// container will be parent+2.
				//
				// Need to set the popup's nesting level to
				// parent+3, so that it gets recalculated after
				// the popup gets recalculated.
				3,
				attached_to::combobox_above_or_below,
				create_combobox_popup,
				style},
		 [&]
		 (const auto &peephole_container)
		 {
			 auto impl=ref<custom_combobox_popup_containerObj
			 ::implObj>::create(peephole_container, style);

			 auto textlist_impl=ref<list_elementObj::implObj>
			 ::create(impl, style);

			 return create_p_t_l_impl_ret_t{impl,
					 ref<peepholed_toplevel_listcontainer_layoutmanager_implObj>
					 ::create(impl,
						  list_element::create
						  (textlist_impl)
						  )
					 };

		 },

		 [&]
		 (const popup_attachedto_info &attachedto_info,
		  const ref<custom_combobox_popup_containerObj::implObj> &impl,
		  const ref<listlayoutmanagerObj::implObj> &layout_impl)
		 {
			 auto container=custom_combobox_popup_container
			 ::create(impl,
				  layout_impl,
				  attachedto_info);

			 popup_containerptr=container;

			 return container;
		 });

        custom_combobox_popup_container popup_container=popup_containerptr;
	ref<listlayoutmanagerObj::implObj>
		popup_listlayoutmanager=combobox_popup->get_layout_impl();

	// We can now start creating the combobox display element, starting
	// with the container where everything goes.
	auto combobox_container_impl=
		ref<custom_combobox_containerObj::implObj>
		::create(parent, popup_container,
			 popup_handler);

	// And the layout manager.
	auto lm=create_impl({combobox_container_impl});

	auto glm=lm->create_gridlayoutmanager();

	// The combo-box's internal grid layout manager will have one row
	// with the current selection element, and the button element.
	auto f=glm->append_row();

	// Invoke the callback to construct the current selection element,
	// which is automatically show()n.

	auto capture_current_selection=
		capturefactory::create(combobox_container_impl);

	auto focusable_selection=selection_factory(capture_current_selection);

	auto current_selection=capture_current_selection->get();

	current_selection->impl
		->set_background_color("combobox_background_color");

	current_selection->show_all();

	f->padding(0);
	f->border("combobox_border");
	f->valign(valign::middle);

	f->created_internally(current_selection);

	// Now for the combo-button.
	f->padding(0);
	f->border("combobox_border");
	f->halign(halign::fill);
	f->valign(valign::fill);

	const auto &[ff, combobox_button]=
		create_combobox_button(f->get_container_impl(),
				       popup_handler);

	ff->elementObj::impl
		->set_background_color("combobox_background_color");

	f->created_internally(ff);

	// Point the popup container to the current selection element and
	// the combo-box button element, so both can be sized based on the
	// size of the combo-box's items.
	popup_container->impl
		->set_current_combobox_selection_element_and_button
		(current_selection);

	auto c=custom_combobox_container::create(combobox_container_impl,
						 lm,
						 focusable_selection,
						 combobox_button,
						 combobox_popup);


	// Install The combo-box popup's list layout manager
	// selection_changed callback, to forward to the
	// combo-box selection_changed callback (must be
	// weakly-captured to avoid a circular reference).

	{
		listimpl_info_t::lock lock{
			popup_listlayoutmanager->list_element_singleton
				->impl->textlist_info};

		lock->selection_changed=
			[=, current_selection=make_weak_capture
			 (current_selection, combobox_popup, lm)]
			(const auto &info)
			{
				auto got=current_selection.get();

				if (!got)
					return;

				auto &[e, combobox_popup, lm]=*got;

				lm->selection_changed.get()
				({ lm->create_public_object(),
						e,
						combobox_popup,
						info});
			};
	}

	auto collector=lookup_collector::create();

	c->elementObj::impl->THREAD->run_as
		([=,
		  selection_search=this->selection_search]
		 (IN_THREAD_ONLY)
		 {
			 // Install:

			 // 1. The popup handler as the combobox container's
			 // attached_popup.

			 combobox_container_impl->data(IN_THREAD)
				 .attached_popup=popup_handler;


			 auto &focusable_element=focusable_selection
				 ->get_impl()->get_focusable_element();

			 // Clear the search string collector buffer when
			 // the current selection focusable gains/loses
			 // input focus (the search starts afresh).

			 focusable_element
				 .on_keyboard_focus
				 (IN_THREAD,
				  [collector]
				  (const auto &ignore)
				  {
					  collector->buffer.clear();
				  });

			 // Install an on_key_event, to collect the typed in
			 // text, and trigger a combo-box search.
			 focusable_element.on_key_event
				 (IN_THREAD, [collector,
					      selection_search,
					      c=make_weak_capture
					      (current_selection, lm)]
				  (const auto &key_event,
				   bool activated,
				   const auto &mcguffin)
				  {
					  bool processed=false;

					  auto got=c.get();

					  if (got)
					  {
						  auto &[current_selection, lm]=
							  *got;

						  processed=collector->process
							  (key_event,
							   activated,
							   selection_search,
							   current_selection,
							   lm->create_public_object(),
							   mcguffin);
					  }
					  return processed;
				  });
		 });

	return c;
}

void custom_comboboxlayoutmanagerObj
::selection_changed(const custom_combobox_selection_changed_t &cb)
{
	impl->selection_changed=cb;
}

ref<custom_comboboxlayoutmanagerObj::implObj
    > new_custom_comboboxlayoutmanager::create_impl(const create_impl_info &i)
	const
{
	return ref<custom_comboboxlayoutmanagerObj::implObj>
		::create(i.container_impl, *this);
}

LIBCXXW_NAMESPACE_END
