/*
** Copyright 2017 Double Precision, Inc.
** See COPYING for distribution information.
*/
#include "libcxxw_config.h"
#include "connection_info.H"
#include "connection_thread.H"
#include "returned_pointer.H"
#include "catch_exceptions.H"
#include "window_handler.H"
#include "batch_queue.H"
#include <x/sysexception.H>
#include <x/functionalrefptr.H>
#include <x/mcguffinmultimap.H>
#include <atomic>

LIBCXXW_NAMESPACE_START

// Received a message to stop, politely.

void connection_threadObj::stop()
{
	stop_politely();
}

void connection_threadObj::dispatch_stop_politely()
{
	stop_received=true;
	LOG_DEBUG("Connection thread stop message received");
}

void connection_threadObj::dispatch_install_on_disconnect(const std::function<void ()> &callback)
{
	disconnect_callback_thread_only=callback;
}

// Figure out what the connection thread needs to do next. It could be:
//
// 1. Dispatch a message from its message queue.
//
// 2. An X event, received via xcb_poll_event.
//
// 3.
LOG_FUNC_SCOPE_DECL(LIBCXX_NAMESPACE::w::connection_threadObj::run_something,
		    runLogger);

bool connection_threadObj
::run_something(msgqueue_auto &msgqueue,
		struct pollfd *topoll,
		size_t &npoll,
		int &poll_for)
{
	connection_thread thread_(this);

	LOG_FUNC_SCOPE(runLogger);

	// Assume we'll poll() indefinitely, unless there's a change in plans.

	auto batch_queue=get_batch_queue();
	bool maybe_theres_something_in_batch_queue=false;

	for ( ; ; maybe_theres_something_in_batch_queue=true)
	{
		poll_for= -1;

		if (recalculate_containers(IN_THREAD))
			continue;

		if (process_element_position_updated(IN_THREAD))
			continue;

		if (process_visibility_updated(IN_THREAD))
			continue;

		// Process a message in the message queue. If processing a
		// message resulted in container recalculation, element
		// position update, or visibility update, that was an
		// app request which must be processed before the next message,
		// another app request, gets processed, so we go back to the top
		// before processing the next message.

		if (!msgqueue->empty())
		{
			msgqueue.event();
			continue;
		}

		// Search: are there any scheduled callbacks?

		if (invoke_scheduled_callbacks(IN_THREAD, poll_for))
			continue;

		expire_incremental_updates(IN_THREAD, poll_for);

		// Check if the connection errored out, if not, check for
		// a message.

		if (npoll == 2 && xcb_connection_has_error(info->conn))
		{
			LOG_FATAL("Connection to the X server has a fatal error");
			npoll=1;
			topoll[1].revents=0;
			disconnected_flag_thread_only=true;
			try {
				disconnect_callback_thread_only();
			} CATCH_EXCEPTIONS;

			for (const auto &handler:*window_handlers_thread_only)
				try {
					handler.second->disconnected(IN_THREAD);
				} CATCH_EXCEPTIONS;
		}

		if (npoll == 2)
		{
			if (auto event=return_pointer(xcb_poll_for_event
						      (info->conn)))
			{
				LOG_TRACE("Processing event "
					  << (int)(event->response_type & ~0x80)
					  << (event->response_type & 0x80
					      ? " (SendEvent)":""));

				run_event(IN_THREAD, event);
				continue;
			}

			if (process_buffered_motion_event(IN_THREAD))
				continue;
		}

		if (maybe_theres_something_in_batch_queue)
			// Give the batch queue an opportunity to flush the
			// batched jobs.
			return false;

		if (redraw_elements(IN_THREAD))
			continue;

		if (run_idle(IN_THREAD))
			return false;
		break;
	}

	if (release_grabs_and_process_buffered_events(IN_THREAD))
		return false;

	xcb_flush(info->conn);
	return true;
}

// Insert a new callback

ref<obj> connection_threadObj
::do_schedule_callback(IN_THREAD_ONLY,
		       const tick_clock_t::duration &timeout,
		       const callback_functional_t &callback)
{
	return (*scheduled_callbacks(IN_THREAD))->insert(tick_clock_t::now()
							 + timeout,
							 callback);
}

// Check scheduled_callbacks() list.

// Returns a bool indicating whether we executed something.
// If false is returned, may update poll_for with the timeout for poll(),
// at which point something might be scheduled, then. Otherwise poll_for
// is left alone (it should be initialized to -1.

bool connection_threadObj::invoke_scheduled_callbacks(IN_THREAD_ONLY,
						      int &poll_for)
{
	bool invoked=false;

	// Scan the scheduled_callbacks list.

	auto b=(*scheduled_callbacks(IN_THREAD))->begin();
	auto e=(*scheduled_callbacks(IN_THREAD))->end();

	while (b != e)
	{
		auto p=b->second.getptr();

		if (p.null())
		{
			++b;
			continue; // Zombie callback.
		}

		// At this point we should check the current
		// system clock, and see if the callback time
		// has elapsed.

		auto now=tick_clock_t::now();

		// And take a closer look
		for ( ; b != e && now >= b->first; ++b)
		{
			if (p.null())
				// Second, and subsequent, time here.
				p=b->second.getptr();

			if (p.null())
				continue; // Ignore zombies

			// Remove this callback, and invoke it.
			b->second.erase();

			try {
				p->invoke(IN_THREAD);
			} CATCH_EXCEPTIONS;
			invoked=true;
			p=nullptr;

			// At this point, if the iterator hasn't reached
			// end, b->first must be greater than now.
		}

		if (b != e)
			poll_for=compute_poll_until(now, b->first);

		break;
	}

	return invoked;
}

int connection_threadObj::compute_poll_until(tick_clock_t::time_point now,
					     tick_clock_t::time_point when)
{
	const auto maximum_poll=
		std::chrono::duration_cast<tick_clock_t::duration>
		(std::chrono::minutes(30));

	// We don't have anything more than
	// a few seconds in the future, but
	// just in case we have long terrm
	// plans, limit poll() to 30 minute
	// timeouts.

	if (when > now + maximum_poll)
		return 30*60 * 1000;

	auto v=std::chrono::duration_cast
		<std::chrono::milliseconds>
		(when-now).count();

	// It's possible if the system
	// clock's precision is greater
	// than a millisecond, the
	// duration cast produces a
	// goose egg. poll() for a lone
	// millisecond, in that case.

	if (v == 0)
		v=1;
	return v;
}

void connection_threadObj
::dispatch_do_run_as(const x::function<void (IN_THREAD_ONLY)> &func)
{
	LOG_FUNC_SCOPE(runLogger);

	// Make sure all changes in the main execution thread are
	// committed by now. Although this should theoretically
	// taken care of by the mutex, this is technically required
	// for the connection thread to see what it needs to see.

	std::atomic_thread_fence(std::memory_order_acquire);

	func(connection_thread(this));
}

bool connection_threadObj::run_idle(IN_THREAD_ONLY)
{
	if (idle_callbacks(IN_THREAD)->empty())
		return false;

	auto f=idle_callbacks(IN_THREAD)->front();
	idle_callbacks(IN_THREAD)->pop_front();

	try {
		f(IN_THREAD);
	} CATCH_EXCEPTIONS;
	return true;
}

LIBCXXW_NAMESPACE_END
