/*
** Copyright 2017 Double Precision, Inc.
** See COPYING for distribution information.
*/
#ifndef focusable_h
#define focusable_h

#include "x/w/focusableobj.H"
#include "x/w/elementobj.H"
#include "focus/focusablefwd.H"
#include "focus/label_forfwd.H"
#include "connection_threadfwd.H"
#include <array>
#include <utility>
#include <x/weaklist.H>

LIBCXXW_NAMESPACE_START

class child_elementObj;

//! Return true if this key event is for advancing focus to the next field.

bool next_key(const key_event &) LIBCXX_HIDDEN;

//! Return true if keypress, and next_key().

bool next_key_pressed(const key_event &) LIBCXX_HIDDEN;

//! Return true if this key event is for advancing focus to the previous field.

bool prev_key(const key_event &) LIBCXX_HIDDEN;

//! Return true if keypress, and prev_key().

bool prev_key_pressed(const key_event &) LIBCXX_HIDDEN;

//! Return true if this key event is for advancing focus to the next page.

bool next_page_key(const key_event &) LIBCXX_HIDDEN;

//! Return true if keypress, and next_page_key().
bool next_page_key_pressed(const key_event &) LIBCXX_HIDDEN;

//! Return true if this key event is for advancing focus to the previous page.

bool prev_page_key(const key_event &) LIBCXX_HIDDEN;

//! Return true if keypress, and prev_page_key().
bool prev_page_key_pressed(const key_event &) LIBCXX_HIDDEN;

//! Return true if this key event is for selecting something.

bool select_key(const key_event &) LIBCXX_HIDDEN;

//! Return true if keypress, and select_key().
bool select_key_pressed(const key_event &) LIBCXX_HIDDEN;

//! Focusable field implementation object.

//! The focusable_elementObj template inherits from this class and a subclass
//! of child_elementObj, and implements get_focusable_element().
//!
//! The public object must inherit from focusableObj::ownerObj, which takes
//! care of invoking focusable_initialize() and focusable_deinitialize().

class LIBCXX_HIDDEN focusableImplObj : virtual public obj {

	//! My iterator in the focusable_fields list.

	//! The public object's constructor and destructor is responsible
	//! for inserting me into the list.
	//!
	//! \note not valid unless in_focusable_fields is set.

	focusable_fields_t::iterator focusable_fields_iter_thread_only;

	//! Whether this object is now fully constructed.
	bool in_focusable_fields_thread_only;

	// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	// NOTE: USE GET_FOCUSABLE_FIELDS_ITER macro to access the above two!!!
	// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	//! Container of all labels for this focusable.
	typedef weaklist<label_forObj> my_labels_t;

	//! All labels for this focusable.

	my_labels_t my_labels_thread_only;

 public:

	THREAD_DATA_ONLY(focusable_fields_iter);
	THREAD_DATA_ONLY(in_focusable_fields);
	THREAD_DATA_ONLY(my_labels);

	//! Candidate for receiving focus automatically when shown.

	mpobj<bool> autofocus;

	//! Constructor
	focusableImplObj();

	//! Destructor
	~focusableImplObj();

	//! Whether this element is enabled for focusing.

	bool focusable_enabled(IN_THREAD_ONLY);

	//! Set enabled/disabled

	void set_enabled(IN_THREAD_ONLY, bool flag);

	//! Must be implement in the subclass.

	//! Returns this focusable's element.
	//! \see focusable_element

	virtual child_elementObj &get_focusable_element()=0;

	//! Must be implement in the subclass.

	//! Returns this focusable's element.
	//! \see focusable_element

	virtual const child_elementObj &get_focusable_element() const=0;

	//! Called from the public object's constructor to finish construction.

	//! Inserts itself into the focusable_fields list.
	void focusable_initialize(IN_THREAD_ONLY);

	//! Invoke from the public object's destructor

	void focusable_deinitialize(IN_THREAD_ONLY);

	//! Return the focusable fields.
	focusable_fields_t &focusable_fields(IN_THREAD_ONLY);

	//! Advance input focus to the next focusable field.
	void next_focus(IN_THREAD_ONLY);

	//! This is a label for this display element.
	void set_label(const ref<elementObj::implObj> &);

 private:
	//! Advance input focus to the next focusable field, on or after this iterator.

	//! The public object's destructor invokes next_focus() after removing
	//! focusable_fields_iter from focusable_fields.

	void next_focus(IN_THREAD_ONLY,
			focusable_fields_t::iterator starting_iter);
 public:
	//! Advance input focus to the previous focusable field.
	void prev_focus(IN_THREAD_ONLY);

	//! Set focus to this focusable element.

	//! Sets the focus, but doesn't bother to call ensure_visibility().
	//! Typically done in response to a button press, because unexpected
	//! scrolling as a result of pressing a button would be a rather rude
	//! thing to do.

	void set_focus_only(IN_THREAD_ONLY);

	//! Set focus to this focusable element and make sure it is visible.

	virtual void set_focus_and_ensure_visibility(IN_THREAD_ONLY);

 private:

	//! Called to move input focus to another focusable field.

	//! Invokes focus_to's set_focus_and_ensure_visibility().

	void switch_focus(IN_THREAD_ONLY, const focusable_impl &focus_to);

 public:
	//! Set tabbing order.
	void get_focus_before(IN_THREAD_ONLY,
			      const ref<focusableImplObj> &other);
	//! Set tabbing order.
	void get_focus_after(IN_THREAD_ONLY,
			      const ref<focusableImplObj> &other);

 private:
	//! Set tabbing order.
	void i_will_get_focus_before(IN_THREAD_ONLY,
				     focusableImplObj &other);
	//! Set tabbing order.
	void i_will_get_focus_after(IN_THREAD_ONLY,
				    focusableImplObj &other);

 public:

	//! Throw an exception

	//! Invoked by the focusable_elementObj mixin when a child focusable
	//! element gets created.
	//!
	//! Throw an exception, can't do it.

	void creating_focusable_child_element();


	//! Make sure we do not have input focus
	void unfocus(IN_THREAD_ONLY);
};

void get_focus_before_in_thread(IN_THREAD_ONLY,	const focusable &me,
				const focusable &other) LIBCXX_HIDDEN;

void get_focus_after_in_thread(IN_THREAD_ONLY,	const focusable &me,
				const focusable &other) LIBCXX_HIDDEN;

void get_focus_impl_after_in_thread(IN_THREAD_ONLY, const focusable &me,
				    const ref<focusableImplObj> &a)
	LIBCXX_HIDDEN;

//! focusableObj::do_get_impl() constructs this object and invokes the
//! callback.

struct LIBCXX_HIDDEN internal_focusable_group {

	//! How many internal focusables there are in this group.
	size_t internal_impl_count;

	//! Pointer to an array of intenral_impl_count focusable implementation objects.
	ref<focusableImplObj> *impls;
};

//! Fixed list of focusable elements in a single group.

//! This is used by an focusableObj::do_get_impl() implementation to
//! invoke its callback, passing it a fixed list of focusable implementation
//! objects.

template<typename ...Args>
inline void process_focusable_impls(const function<internal_focusable_cb> &cb,
				    Args && ...args) LIBCXX_HIDDEN;

template<typename ...Args>
inline void process_focusable_impls(const function<internal_focusable_cb> &cb,
				    Args && ...args)
{
	std::array<ref<focusableImplObj>, sizeof...(Args)> impls={
		std::forward<Args>(args)...};

	internal_focusable_group group{impls.size(), &impls[0]};

	cb(group);
}

//! Take a vector of focusables, extract their implementations, invoke the callback.

//! Note, this individually grabs all impls from do_get_impl(), and returns
//! from it. Handle with care. Make sure all locks are in place.

void process_focusable_impls_from_focusables
(const function<internal_focusable_cb> &cb,
 const std::vector<focusable> &v)
	LIBCXX_HIDDEN;

LIBCXXW_NAMESPACE_END

#endif
